from yta_editor_frame.frame.resource.abstract import FrameResourceABC
from yta_editor_frame.frame.resource.types import FrameResourceType
from yta_programming.decorators.requires_dependency import requires_dependency
from yta_validation.parameter import ParameterValidator


class AvVideoFrameResource(
    FrameResourceABC
):
    """
    A frame resource based on a pyav VideoFrame.

    The `.frame` attribute is an `av.VideoFrame`.
    """

    @property
    def width(
        self
    ) -> int:
        return self.frame.width
    
    @property
    def height(
        self
    ) -> int:
        return self.frame.height
    
    @property
    def format(
        self
    ) -> str:
        return str(self.frame.format.name)
    
    @property
    def type(
        self
    ) -> FrameResourceType:
        return FrameResourceType.CPU_AV_VIDEOFRAME

    def __init__(
        self,
        frame: 'av.VideoFrame',
        render_context: 'RenderContext'
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, 'VideoFrame')

        super().__init__(
            frame = frame,
            render_context = render_context
        )

    def release(
        self
    ):
        pass

    @requires_dependency('av', 'yta_editor_frame', 'av')
    def copy(
        self
    ) -> 'AvVideoFrameResource':
        """
        Get a copy of this frame resource.
        """
        import av

        arr = self.frame.to_ndarray()
        arr_copy = arr.copy()
        # TODO: Why do we force the 'format' (?)
        copy = av.VideoFrame.from_ndarray(arr_copy, format = 'rgb24')

        return AvVideoFrameResource(
            frame = copy,
            render_context = self.render_context
        )


from yta_validation.parameter import ParameterValidator
from abc import abstractmethod, ABC


class FrameResourceABC(ABC):
    """
    Abstract method to represent a frame resource.
    """

    @property
    @abstractmethod
    def width(
        self
    ) -> int:
        """
        The width of the frame resource, in pixels.
        """
        pass

    @property
    @abstractmethod
    def height(
        self
    ) -> int:
        """
        The height of the frame resource, in pixels.
        """
        pass

    @property
    @abstractmethod
    def format(
        self
    ) -> int:
        """
        The format of the frame resource.

        TODO: Is this useful (?)
        """
        pass

    @property
    @abstractmethod
    def type(
        self
    ) -> 'FrameResourceType':
        """
        The type of frame resource.
        """
        pass

    def __init__(
        self,
        # TODO: Set the types
        frame: any,
        render_context: 'RenderContext'
    ):
        ParameterValidator.validate_mandatory_instance_of('render_context', render_context, 'RenderContext')

        # TODO: Set the types
        self.frame: any = frame
        """
        The frame itself.
        """
        self.render_context: 'RenderContext' = render_context
        """
        *For internal use only*

        The `RenderContext` instance, useful to release
        the resources when needed.
        """

    @abstractmethod
    def release(
        self
    ):
        """
        Release the resource, if needed.
        """
        raise NotImplementedError
    
    @abstractmethod
    def copy(
        self
    ):
        """
        Get a copy of this frame resource.
        """
        raise NotImplementedError


from yta_editor_frame.frame.resource.abstract import FrameResourceABC
from yta_editor_frame.frame.resource.types import FrameResourceType
from yta_editor_frame.utils import copy_texture
from yta_validation.parameter import ParameterValidator


class GPUTextureFrameResource(
    FrameResourceABC
):
    """
    A frame resource based on a texture.

    The `.frame` attribute is a `moderngl.Texture`.
    """
    
    @property
    def width(
        self
    ) -> int:
        return self.frame.width
    
    @property
    def height(
        self
    ) -> int:
        return self.frame.height
    
    @property
    def format(
        self
    ) -> str:
        return self.frame.dtype
    
    @property
    def type(
        self
    ) -> FrameResourceType:
        return FrameResourceType.GPU

    def __init__(
        self,
        frame: 'moderngl.Texture',
        render_context: 'RenderContext'
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, 'Texture')

        super().__init__(
            frame = frame,
            render_context = render_context
        )

    def release(
        self
    ):
        self.render_context.gpu.texture.release(self.frame)

    def copy(
        self
    ) -> 'GPUTextureFrameResource':
        """
        Get a copy of this frame resource.
        """
        copy = copy_texture(
            texture = self.frame,
            render_context = self.render_context,
            do_get_from_pool = True
        )

        return GPUTextureFrameResource(
            frame = copy,
            render_context = self.render_context
        )
        